package kosoft.testproxy;

//Copyright (C) 2000 Mark O'Donohue [ mark.odonohue@gmail.com] 
//All rights reserved.
//
//This library is free software; you can redistribute it and/or modify it under the terms of the GNU 
//Lesser General Public License as published by the Free Software Foundation; either version 2.1 of 
//the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
//even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//Lesser General Public License for more details.


import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.StringTokenizer;


public class MonitorControl implements Service {
	ProxyServer server;             // The server we control
	String password;           // The password we require
	boolean connected = false; // Whether a client is already connected to us
	/**
	 * Create a new Control service.  It will control the specified Server
	 * object, and will require the specified password for authorization
	 * Note that this Service does not have a no argument constructor, which
	 * means that it cannot be dynamically instantiated and added as the other,
	 * generic services above can be.
	 **/
	public MonitorControl(ProxyServer server, String password) {
		this.server = server;
		this.password = password;
	}

	void sourceFile(String fileName, PrintWriter out) {
		try { 
			BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(fileName)));
			serveCommands(true, in, out);
			in.close();
		}
		catch (FileNotFoundException e) {
			out.println("File Not Found " + fileName);
		}
		catch (IOException e) {
			out.println("IO Error reading file " + fileName);
		}
	}

	/**
	 * This is the serve method that provides the service.  It reads a line 
	 * the client, and uses java.util.StringTokenizer to parse it into
	 * commands and arguments.  It does various things depending on the 
	 * command.
	 **/
	public void serve(InputStream i, OutputStream o) throws IOException {
		// Setup the streams
		BufferedReader in = new BufferedReader(new InputStreamReader(i));
		PrintWriter out = new PrintWriter(new OutputStreamWriter(o));

		// If there is already a client connected to this service, display a
		// message to this client and close the connection.  We use a 
		// synchronized block to prevent a race condition.


		synchronized(this) { 
			if (connected) { 
				out.println("ONLY ONE CONTROL CONNECTION ALLOWED AT A TIME.");
				out.close();
				return;
			}
			else connected = true;
		}

		// Check if we have an init file.  
		String initFile = "proxyinit.txt";
		File file = new File(initFile);
		if (file.exists()) {
			System.out.println("Run init file :" + initFile);
			sourceFile(initFile, out);
		}
		else {
			System.out.println("No init file '" + initFile + "' found");
		}

		// Check if we are to not look for input
		if (System.getProperty("runBackground") != null) {
			cmdRunBackground(out);
		}

		serveCommands(false, in, out);

		// Finally, when the loop command loop ends, close the streams
		// and set our connected flag to false so that other clients can
		// now connect.
		//if (out != System.out) out.close();
		//if (in != System.in)  in.close();

		out.close();
		in.close();

		connected = false;

	}


	public void serveCommands(boolean echo, BufferedReader in, PrintWriter out) throws IOException {

		String line;

		//boolean authorized = false;  // Has the user has given the password yet?
		boolean authorized = true;  // Has the user has given the password yet?
		int num;

		for(;;) {  // infinite loop
			out.print("> ");           // Display a prompt
			out.flush();               // Make it appear right away
			line = in.readLine();      // Get the user's input
			
			
			if (line == null) {
				break;   // Quit if we get EOF.
			}
			
			if (echo) {
				out.println(line);
			}
			
			line = line.trim();
			
			if (line.length() == 0 || line.startsWith("#")) {  // skip comments
				continue;
			}
			
			int hashIdx =line.indexOf('#');   // Trim out trainling comments on line
			if ( hashIdx != -1) {
				line = line.substring(0, hashIdx);
				line = line.trim();
			}
			
			try {
				// Use a StringTokenizer to parse the user's command
				StringTokenizer t = new StringTokenizer(line);
				if (!t.hasMoreTokens()) continue;  // if input was blank line
				// Get the first word of the input and convert to lower case
				//String command = t.nextToken().toLowerCase(); 
				String command = t.nextToken(); 
				// Now compare it to each of the possible commands, doing the
				// appropriate thing for each command
				
				// If not authorized then only cmd available is password.
				if (!authorized) {
					if (command.equals("password")) {       // Password command
						String p = t.nextToken();             // Get the next word of input
						if (p.equals(this.password)) {        // Does it equal the password
							out.println("OK");                  // Say so
							authorized = true;                  // Grant authorization
						}
						else out.println("INVALID PASSWORD"); // Otherwise fail
					}
					continue;
				}

				// The normal user commands.
				
				if (command.equals("createProxy")) {       
					String params = t.nextToken();
					cmdCreateProxyService(params);
				}
				else if (command.equals("removeProxy")) {
					int port = Integer.parseInt(t.nextToken());
					cmdRemoveProxyService(port);
					out.println("SERVICE REMOVED");
				}
				else if (command.equals("startProxy")) {
					int port = Integer.parseInt(t.nextToken());
					cmdStartProxyService(port);
					out.println("SERVICE STARTED");
				}
				else if (command.equals("stopProxy")) {
					int port = Integer.parseInt(t.nextToken());
					cmdStopProxyService(port);
					out.println("CONNECTIONS CLOSED AND SERVICE STOPPED");
				}
				else if (command.equals("stopProxyListen")) {
					int port = Integer.parseInt(t.nextToken());
					cmdStopProxyListenService(port);
					out.println("SERVICE STOPPED");
				}
				else if (command.equals("startDownIP")) {
					int port = Integer.parseInt(t.nextToken());
					cmdStartDownIPForProxy(port);
					out.println("IP DOWN for Proxy");
				}
				else if (command.equals("stopDownIP")) {
					int port = Integer.parseInt(t.nextToken());
					cmdStopDownIPForProxy(port);
					out.println("IP UP for Proxy");
				}
				else if (command.equals("startProxyLog")) {
					int port = Integer.parseInt(t.nextToken());
					cmdStartProxyLog(port);
					out.println("SERVICE STARTED");
				}
				else if (command.equals("startProxyLogToFile")) {
					int port = Integer.parseInt(t.nextToken());
					cmdStartProxyLogToFile(port);
					out.println("Proxy Log to File STARTED");
				}
				else if (command.equals("stopProxyLog")) {
					int port = Integer.parseInt(t.nextToken());
					cmdStopProxyLog(port);
					out.println("CONNECTIONS CLOSED AND SERVICE STOPPED");
				}
				
				else if (command.equals(".")) {       // Run commands from file
					String params = t.nextToken();
					sourceFile(params, out);
				}
				else if (command.equals("source")) {       // Run commands from file
					String params = t.nextToken();
					sourceFile(params, out);
				}
				else if (command.equals("close")) {       // Add Service command
					int localport = Integer.parseInt(t.nextToken());
					int conNum = Integer.parseInt(t.nextToken());

					cmdCloseConnection(localport, conNum);
				}
				else if (command.equals("closeAll")) {       // Add Service command
					int localport = Integer.parseInt(t.nextToken());

					cmdCloseAll(localport);
				}
				else if (command.equals("sleep")) {
					long timeMs = Integer.parseInt(t.nextToken()) * 1000;
					cmdSleep(timeMs);
				}
				else if (command.equals("max")) {      // Set max connection limit
					int max = Integer.parseInt(t.nextToken());        // get limit
					server.connectionManager.setMaxConnections(max);  // set limit
					out.println("MAX CONNECTIONS CHANGED");           // acknowledge
				}
				else if (command.equals("status")) {    // Status Display command
					
					boolean printConnections = false; 
					if (t.hasMoreElements()) { 
						int localport = Integer.parseInt(t.nextToken());
						cmdPrintStatus(localport, out, printConnections);
					}
					else { 
						cmdPrintStatus(out, printConnections);
					}
				}
				else if (command.equals("statusFull")) {    // Status Display command
					
					boolean printConnections = true; 
					if (t.hasMoreElements()) { 
						int localport = Integer.parseInt(t.nextToken());
						cmdPrintStatus(localport, out, printConnections);
					}
					else { 
						cmdPrintStatus(out, printConnections);
					}
				}
				else if (command.equalsIgnoreCase("setSendDelay")) {      // 
					int localport = Integer.parseInt(t.nextToken());        // get limit
					int delay = Integer.parseInt(t.nextToken());        // get limit

					cmdSetSendDelay(localport, delay);
					out.println("SendDelay CHANGED");           // acknowledge
				}
				else if (command.equalsIgnoreCase("setRecvDelay")) {      // 
					int localport = Integer.parseInt(t.nextToken());        
					int delay = Integer.parseInt(t.nextToken());        

					cmdSetRecvDelay(localport, delay);
					out.println("RecvDelay CHANGED");           // acknowledge
				}
				else if (command.equalsIgnoreCase("setSendDelayRange")) {      // 
					int localport = Integer.parseInt(t.nextToken());        // get limit
					int delay = Integer.parseInt(t.nextToken());        // get limit

					cmdSetSendDelayRange(localport, delay);
					out.println("SendDelayRange CHANGED");           // acknowledge
				}
				else if (command.equalsIgnoreCase("setRecvDelayRange")) {      // 
					int localport = Integer.parseInt(t.nextToken());        
					int delay = Integer.parseInt(t.nextToken());        

					cmdSetRecvDelayRange(localport, delay);
					out.println("RecvDelayRange CHANGED");           // acknowledge
				}
				else if (command.equalsIgnoreCase("startDegrade")) {      // 
					int localport = Integer.parseInt(t.nextToken());        
					int incTimeSec = Integer.parseInt(t.nextToken());        
					int incDelay = Integer.parseInt(t.nextToken());        

					cmdStartDegrade(localport, incTimeSec, incDelay);
					out.println("startDegrade");           // acknowledge
				}
				else if (command.equalsIgnoreCase("stopDegrade")) {      // 
					int localport = Integer.parseInt(t.nextToken());        
					cmdStopDegrade(localport);
					out.println("stopDegrade");           // acknowledge
				}
				else if (command.equalsIgnoreCase("setRandomFailPerc")) {      // 
					int localport = Integer.parseInt(t.nextToken());
					double failPerc = Double.parseDouble(t.nextToken());        
					cmdSetRandomFailPerc(localport, failPerc);
				}
				else if (command.equalsIgnoreCase("startRandomFail")) {      // 
					int localport = Integer.parseInt(t.nextToken());
					if (t.hasMoreTokens()) {
						double failPerc = Integer.parseInt(t.nextToken());        
						cmdStartRandomFail(localport, failPerc);
					}
					else {
						cmdStartRandomFail(localport);						
					}
					out.println("startRandomFail");           // acknowledge
				}
				else if (command.equalsIgnoreCase("stopRandomFail")) {      // 
					int localport = Integer.parseInt(t.nextToken());
					cmdStopRandomFail(localport);						
					out.println("stopRandomFail");           // acknowledge
				}
				else if (command.equalsIgnoreCase("startAutoDelay")) {      // 
					int localport = Integer.parseInt(t.nextToken());

					// If we have more tokens then parse the settings. 
					if (t.hasMoreTokens()) {
						int delayTime = Integer.parseInt(t.nextToken());

						int cycleWaitTime = -1;
						int activeDelayTime = -1;  
						
						if (t.hasMoreTokens()) {
							cycleWaitTime = Integer.parseInt(t.nextToken());
							activeDelayTime = cycleWaitTime; // default to the same 
							if (t.hasMoreTokens()) {
								activeDelayTime = Integer.parseInt(t.nextToken());						
							}
						}
						cmdSetAutoDelaySchedulerSpec(localport, delayTime, cycleWaitTime, activeDelayTime);
					}
					
					cmdStartAutoDelayScheduler(localport);
					out.println("startAutoDelay");           // acknowledge
				}
				else if (command.equalsIgnoreCase("stopAutoDelay")) {      // 
					int localport = Integer.parseInt(t.nextToken());        
					cmdStopAutoDelayScheduler(localport);
					out.println("stopAutoDelay");           // acknowledge
				}
				else if (command.equalsIgnoreCase("setAutoDelaySpec")) {      // 
					int localport = Integer.parseInt(t.nextToken());

					int delayTime = Integer.parseInt(t.nextToken());

					int cycleWaitTime = -1;
					int activeDelayTime = -1;  
					
					if (t.hasMoreTokens()) {
						cycleWaitTime = Integer.parseInt(t.nextToken());
						activeDelayTime = cycleWaitTime; // default to the same 
						if (t.hasMoreTokens()) {
							activeDelayTime = Integer.parseInt(t.nextToken());						
						}
					}
					cmdSetAutoDelaySchedulerSpec(localport, delayTime, cycleWaitTime, activeDelayTime);
					out.println("setAutoDelaySpec");           // acknowledge
				}
				else if (command.equalsIgnoreCase("startSilent")) {      // 

					cmdStartSilent();
					out.println("startSilent");           // acknowledge
				}
				else if (command.equalsIgnoreCase("stopSilent")) {      // 

					cmdStopSilent();
					out.println("stopSilent");           // acknowledge
				}
				else if (command.equalsIgnoreCase("rmOldLogFiles")) {      // 
					cmdRmOldLogFiles();
					out.println("rmOldLogFIles");           // acknowledge
				}
				else if (command.equals("help")) {            // Help command
					// Display command syntax.  Password not required
					cmdHelp(out);

				}
				else if (command.equals("runBackground")) {
					cmdRunBackground(out);
				}
				else if (command.equals("version")) {            // Help command
					// Display command syntax.  Password not required
					cmdPrintVersion(out);

				}
				else if (command.equals("quit")) break;    // Quit command.  Exit.
				else out.println("UNRECOGNIZED COMMAND");  // Unknown command error
			}
			catch (Exception e) {
				// If an exception occurred during the command, print an error
				// message, then output details of the exception.
				out.println("EXCEPTION WHILE PARSING OR EXECUTING COMMAND:");
				out.println(e);
				e.printStackTrace(out);
			}
		}
	}

	void cmdSleep(long time) {
		try { 
			Thread.sleep(time);
		}
		catch (InterruptedException e) {} 
	}
	
	void cmdRunBackground(PrintWriter out) {
		
		cmdPrintStatus(out, false);
		out.println("sleeping in background indefinately...." );
		out.flush();
		while (true) { 
			try { 
				Thread.sleep(30000);
			}
			catch (InterruptedException e) {}
		}
		
	}
	
	void cmdPrintVersion(PrintWriter out) {
		out.println(ProxyServer.version);
	}
	
	
	void cmdHelp(PrintWriter out) {
		out.println(				
				ProxyServer.version  + "\n COMMANDS:\n" + 
				//"\tpassword <password>\n" +
				"\tcreateProxy <localport>:<remotehost>:<remoteport>\t- Create new proxy service\n" +
				"\tremoveProxy <localport>\t- remove proxy service, and close all client connections\n" +
				"\n" + 
				"\tstartProxy <localport>\t- Start listenin on localport for new connections (need previously setup service)\n" +
				"\tstopProxy <localport>\t- Stop listening for new connection, and close existing connections\n" +
				"\tstopProxyListen <localport>\t- Stop listening for new connection, but leave existing ones open\n" +
				"\n" + 
				"\tstartDownIP <localport>\t- simulate remote IP address down, here we drop all send packets\n" +
				"\tstopDownIP <localport>\t- bring remote IP backonline, new requests will then go through (old ones are dropped)\n" +
				"\tmax <max-connections>\n" +
				"\n" + 
				"\tsetRandomFailPerc <localport> <percfail%>\t- Set random fail percent value from 0.0-1.0\n" + 
				"\tstartRandomFail <localport> [<percfail%>]\t- Start randomly failing send data (perc: from 0.0-1.0)\n" + 
				"\tstopRandomFail <localport>\t- Stop Random fail\n" + 
				"\n" + 
				"\tsetSendDelay <localport> <ms>\t- Add a delay in client requests before send to backend\n" + 
				"\tsetRecvDelay <localport> <ms>- Add a delay in client requests before pass back to client\n" + 
				"\n" + 
				"\tsetSendDelayRange <localport> <ms>\t- Add a range, based on mid point of the abive send delay\n" + 
				"\tsetRecvDelayRange <localport> <ms>- Add a range, based on mid point of the abive send delay\n" + 
				"\n" + 
				"\tsetAutoDelaySpec <localport> <delayms> [<waitms> <applyms>] \t- Set the delay, the wait time and the apply time (default 2min, 120000ms)\n" + 
				"\tstartAutoDelay <localport> [<delayms> [<waitms> <applyms>]]\t- Cycle with delay for 2min, then clear for 2min repeat\n" + 
				"\tstopAutoDelay <localport> - Stop auto delay cycle\n" + 
				"\n" + 

				"\tstartDegrade <localport> <incTimeSec> <incDelayMs>\t- Incrementally degrade service, every incTimeSec we add incDelayMs to the recvDelay\n" + 
				"\tstopDegrade <localport>\t- Stop incremental delay\n" + 
				"\n" + 
				"\tstartProxyLog <localport>\t- print log of data through proxy\n" + 
				"\tstartProxyLogToFile <localport>\t- trace proxy mesages to named log file in log subdirectory\n" + 
				"\tstopProxyLog <localport>\t- Stop logging of server\n" + 
				"\n" + 
				"\tclose <localport> <num>\t- Close a specific client connection from 'status' report\n" + 
				"\tcloseAll <localport>\t- Close all client connections for localport (listener remains active for new clients however)\n" + 
				"\n" + 
				"\tstatus <localport>\t-print status of one port\n" +
				"\tstatusFull <localport>\t-print status plus connecion list\n" +
				"\tstatus\t-print status of all ports\n" +
				"\tstatusFull\t-print status plus connection list\n" +
				"\n" + 
				"\trunBackground\t- run in background - does not require input\n" +
				"\t\t  used if need to run from nohup script,there is also a runBatch.sh \n" +
				"\t\t  which sets a cmdline switch t \n" +
				"\n" + 
				"\tstartSilent\t- stop displaying messages from new connect and closed connections\n" +
				"\t\t  (they can overwealm the display at times)\n" +
				"\tstopSilent\t- start displaying messages from new connect and closed connections\n" +
				"\n" + 
				"\t. <filename>\t- take commands from file\n" +
				"\tsource <filename>\t-takes commands from file\n" +
				"\tsleep <time-sec>\t-sleep for time (for use in script file)\n" +
				"\trmOldLogFiles\t- remove old log files from log sub directory\n" +
				"\tversion\t- print version of ProxyTool\n" +
				"\n" + 
		"\tquit");

	}



	void cmdSetSendDelay(int localport, long delay) {

		ProxyService service = server.requireProxy(localport);
		service.setSendDelay(delay);
		
	}

	void cmdSetRecvDelay(int localport, long delay) {

		ProxyService service = server.requireProxy(localport);
		service.setRecvDelay(delay);
		
	}


	void cmdSetSendDelayRange(int localport, long delay) {

		ProxyService service = server.requireProxy(localport);
		service.setSendDelayRange(delay);
		
	}
	
	void cmdSetRecvDelayRange(int localport, long delay) {

		ProxyService service = server.requireProxy(localport);
		service.setRecvDelayRange(delay);		
	}

	
	void cmdStartAutoDelayScheduler(int localport) {

		ProxyService service = server.requireProxy(localport);
		service.startAutoDelayScheduler();
		
	}

	void cmdStopAutoDelayScheduler(int localport) {

		ProxyService service = server.requireProxy(localport);
		service.stopAutoDelayScheduler();
		
	}

	void cmdSetAutoDelaySchedulerSpec(int localport, long delayTime, long cycleWaitTime, long activeDelayTime) {

		ProxyService service = server.requireProxy(localport);
		service.setAutoDelaySchedulerDelayTime(delayTime);
		service.setAutoDelaySchedulerCycleTime(cycleWaitTime);
		service.setAutoDelaySchedulerActiveDelay(activeDelayTime);
		
	}
	
	
	

	void cmdStartDegrade(int localport, long incTimeSec, long incDelay) {

		ProxyService service = server.requireProxy(localport);
		service.startIncDelay(incTimeSec, incDelay);
		
	}

	void cmdStopDegrade(int localport) {

		ProxyService service = server.requireProxy(localport);
		service.stopIncDelay();
		
	}

	void cmdSetRandomFailPerc(int localport, double failPerc) {
		ProxyService service = server.requireProxy(localport);
		service.setRandomFailPerc(failPerc);
	}
	
	void cmdStartRandomFail(int localport, double failPerc) {
		ProxyService service = server.requireProxy(localport);
		service.startRandomFail(failPerc);
	}

	void cmdStartRandomFail(int localport) {
		ProxyService service = server.requireProxy(localport);
		service.startRandomFail();
	}
	
	void cmdStopRandomFail(int localport) {
		ProxyService service = server.requireProxy(localport);
		service.stopRandomFail();
	}
	

	void cmdStartSilent() {
		server.setSilentMode(true);
	}
	
	void cmdStopSilent() {
		server.setSilentMode(false);
	}
	
	
	void cmdPrintStatus(PrintWriter out, boolean printConnections) { 
		// Display a list of all services currently running

		for (ProxyService service : server.services.values()) {
			service.print(out);
			if (printConnections) {
				service.printConnections(out);				
			}
		}
		//for (Integer port : server.listeners.keySet() ) {
		//	ProxyListener listener = server.listeners.get(port);
		//	out.println("SERVICE " + listener.service.getName()+ " ON PORT " + port);
		//	listener.service.printConnections(out);
		//}

		// Display a list of all current connections
		//server.connectionManager.printConnections(out);

		out.println("MAX CONNECTIONS: " + server.connectionManager.maxConnections);
		out.println("Silent Mode    : " + server.isSilentMode());

	}

	void cmdPrintStatus(int localport, PrintWriter out, boolean printConnections) { 

		ProxyListener listener = server.listeners.get(localport);

		if (listener == null) { 
			out.println("SERVICE " + localport + " not found");
			return;
		}

		listener.service.print(out);

		if (printConnections) {
			out.println("SERVICE " + listener.service.getName()+ " ON PORT " + localport);
			listener.service.printConnections(out);
		}
	}

	void cmdCloseConnection(int localport, int num) {
		ProxyService service = server.requireProxy(localport);
		service.closeConnection(num);
	}

	void cmdCloseAll(int localport) {
		ProxyService service = server.requireProxy(localport);
		service.closeAll();
	}

	void cmdCreateProxyService(String params) {

		String paramList[] = params.split(":");

		if (paramList.length != 3) {
			System.err.println("Unable to parse  :  <localport>:<remotehost>:<remoteport> from : " + params);
			return;
		}

		try { 
			int localport = Integer.parseInt(paramList[0]);
			String host = paramList[1];
			int remoteport = Integer.parseInt(paramList[2]);

			ProxyService proxy = server.getProxy(localport);
			if (proxy != null) {
				throw new IllegalArgumentException("Service " + proxy.getName() + " already exists for port " + localport);
			}
			proxy = new ProxyService(server, localport, host, remoteport);
			server.addService(proxy);
		}
		catch (Exception e) {  // Print an error message if anything goes wrong.
			System.err.println(e);
			System.err.println("Unable to parse  :  <localport>:<remotehost>:<remoteport> from : " + params);
		}

	}
	
	void cmdRemoveProxyService(int port) {
		server.removeService(port);			
	}

	void cmdStartProxyService(int port) {
		server.startService(port);			
	}
		
	void cmdStopProxyService(int port) {
		server.stopService(port);
		ProxyService proxy = server.getProxy(port);
		proxy.closeAll();
	}

	void cmdStopProxyListenService(int port) {
		server.stopService(port);			
	}
	
	void cmdStartDownIPForProxy(int port) {
		ProxyService proxy = server.requireProxy(port);
		proxy.startDownIP();
				
	}
		
	void cmdStopDownIPForProxy(int port) {
		ProxyService proxy = server.requireProxy(port);
		proxy.stopDownIP();
	}

	
	void cmdStartProxyLog(int port) {
		ProxyService proxy = server.requireProxy(port);
		proxy.startLog();
				
	}

	void cmdStartProxyLogToFile(int port) {
		
		if (server.checkForOldLogFiles()) {
			System.err.println("There are old log files - please run rmOldLogFiles command");
			return;
		}
		ProxyService proxy = server.requireProxy(port);
		proxy.startLogToFile();
				
	}
	
	void cmdStopProxyLog(int port) {
		ProxyService proxy = server.requireProxy(port);
		proxy.stopLog();
	}

	void cmdRmOldLogFiles() {
		server.removeOldLogFiles();
	}
	
	
}
