package kosoft.testproxy;

//Copyright (C) 2000 Mark O'Donohue [ mark.odonohue@gmail.com] 
//All rights reserved.
//
//This library is free software; you can redistribute it and/or modify it under the terms of the GNU 
//Lesser General Public License as published by the Free Software Foundation; either version 2.1 of 
//the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
//even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//Lesser General Public License for more details.


import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.Date;

public class ProxyConnection extends Thread {
	Socket client;     // The socket to talk to the client through
	ProxyService service;   // The service being provided to that client

	ProxyServer proxyServer = null;
	
	long clientConnectNum = -1;
	long sentByteCount = 0;
	long recvByteCount = 0;

	public long getSendByteCount() { return sentByteCount; } 
	public long getRecvByteCount() { return recvByteCount; } 
	
	
	public long getClientConnectNum() { return clientConnectNum; }
	public void setClientConnectNum(long clientConnectNum) { this.clientConnectNum = clientConnectNum; }
	
	boolean logData = false;
	
	public boolean getLogData() { return logData; }
	public void setLogData(boolean logData) { this.logData = logData; }
	
	/**
	 * This constructor just saves some state and calls the superclass
	 * constructor to create a thread to handle the connection.  Connection
	 * objects are created by Listener threads.  These threads are part of
	 * the server's ThreadGroup, so all Connection threads are part of that
	 * group, too.
	 **/
	public ProxyConnection(ProxyServer proxyServer, Socket client, ProxyService service) {
		super("Server.Connection:" + client.getInetAddress().getHostAddress() +
				":" + client.getPort());
		this.proxyServer = proxyServer;
		this.client = client;
		this.service = service;

		service.addConnection(this);
	}

	public void run() {
		try { 
			InputStream in = client.getInputStream();
			OutputStream out = client.getOutputStream();
			serve(in, out);
			//service.serve(in, out);
		} 
		catch (IOException e) {
			proxyServer.log(e);
		}
		finally {
			service.endConnection(this);
			proxyServer.getConnectionManager().endConnection(); 
		}
	}

	Thread c2s = null;
	Thread s2c = null;
	Socket server = null;

	
	public void serve(InputStream in, OutputStream out) {
		// These are some sockets we'll use.  They are final so they can be used
		// by the anonymous classes defined below.
		final InputStream from_client = in;
		final OutputStream to_client = out;
		final InputStream from_server;
		final OutputStream to_server;

		// Try to establish a connection to the specified server and port
		// and get sockets to talk to it.  Tell our client if we fail.
		try { 
			server = new Socket(service.host, service.port); 
			from_server = server.getInputStream();
			to_server = server.getOutputStream();
		}
		catch (Exception e) {
			PrintWriter pw = new PrintWriter(new OutputStreamWriter(out));
			pw.println("Proxy server could not connect to " + service.host + ":" + service.port);
			pw.flush();
			pw.close();
			try { in.close(); } catch (IOException ex) {}
			return;
		}

		// Create an array to hold two Threads.  It is declared final so that
		// it can be used by the anonymous classes below.  We use an array
		// instead of two variables because given the structure of this program
		// two variables would not work if declared final.

		final Thread[] threads = new Thread[2];
		
		// Define and create a thread to transmit bytes from client to server
		c2s = new Thread() {
			public void run() {
				byte[] buffer = new byte[2048];
				int bytes_read;
				try {
					
					String sendPrefix = getLogPrefix() + " >> ";
					while((bytes_read = from_client.read(buffer)) != -1) {

						service.addSendDelay();
						sentByteCount += bytes_read;

						if (service.isLogData()) {
							logBuffer(sendPrefix, buffer, 0, bytes_read);
						}

						// If we are simulating a down remote IP address, we just drop the packets 
						if (service.isSimilateDownIP()) {
							continue;
						}
						
						if (service.isRandomFail()) {
							double range = service.getRandomFailPerc();
							if (checkIfRandomFail(range)) {							
								continue;
							}
						}
						to_server.write(buffer, 0, bytes_read);
						to_server.flush();
					}
				}
				catch (IOException e) {}

				// if the client closed its stream to us, we close our stream
				// to the server.  First, stop the other thread
				threads[1].stop();
				try { to_server.close(); } catch (IOException e) {}
			}
		};

		// Define and create a thread to copy bytes from server to client.
		s2c = new Thread() {
			public void run() {
				byte[] buffer = new byte[2048];
				int bytes_read;
				try {
					String recvPrefix = getLogPrefix() + " << ";
					while((bytes_read = from_server.read(buffer)) != -1) {

						service.addReplyDelay();
						recvByteCount += bytes_read;

						if (service.isLogData()) {
							logBuffer(recvPrefix, buffer, 0, bytes_read);
						}
						
						// If we are simulating a down remote IP address, we just drop the packets 
						if (service.isSimilateDownIP()) {
							continue;
						}
						to_client.write(buffer, 0, bytes_read);
						to_client.flush();
					}
				}
				catch (IOException e) {}

				// if the server closed its stream to us, we close our stream
				// to the client.  First, stop the other thread, though.
				threads[0].stop();
				try { to_client.close(); } catch (IOException e) {}
			}
		};

		// Store the threads into the final threads[] array, so that the 
		// anonymous classes can refer to each other.
		threads[0] = c2s; threads[1] = s2c;

		// start the threads
		c2s.start(); s2c.start();

		// Wait for them to exit
		try { c2s.join(); s2c.join(); } catch (InterruptedException e) {
		}

		// If we are logging then close that 
		if (logStream != null) {
			logStream.flush();
			logStream.close();
			logStream = null;
		}
		
	}

	static boolean checkIfRandomFail(double range) {
		
		double test = Math.random();
		
		if (test < range) {
			return true;
		}
		return false;
		
	}
	
	void closeConnection() {

		try {
			if (server != null) {
				server.close();
			}
		} catch (IOException e) {
		}
	}
	
	String prefix = null;
	public String getLogPrefix() { 
		if (prefix == null) {
			prefix = client.getLocalPort() + "#" + getClientConnectNum();
		}
		return prefix;
	}
	
	public String toString() {
		String result ="  #" + getClientConnectNum() + " CONNECTED TO " +  client.getInetAddress().getHostAddress() + ":";
		result +=  client.getPort() + " ON PORT " + client.getLocalPort();
		result += " (Sent:" + sentByteCount +", Recv:" + recvByteCount + ") "; 
		result +=  " FOR PROXY SERVICE TO " + service.getDestName();
		
		return result;
		
	}
	
    // Copied these methods to allow printing of the hex and ascii values for the wire traces.
    
    
	PrintWriter logStream = null; 
	
	File getLogFile() {
		String name = "smtestproxy_" + getLogPrefix();
		String fileName = proxyServer.getLogDir() + "/" + name + ".log";
		File file = new File(fileName);
		return file;
		
	}
	public void trace(String line) {
		
		if (service.isLogDataToFile()) {
			if (logStream == null) {
				try {
					logStream = new PrintWriter(new FileWriter(getLogFile()));
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
			}
			logStream.println("[" + new Date() + "] " + line);
			logStream.flush();
		}
		else {
			proxyServer.logWithSilent(line);
		}
	}
    /**
     * Write buffer contents to log file.
     */
    public void logBuffer(String prefix, byte[] b) {
        int bufflen = (b == null) ? 0 : b.length;
        logBuffer(prefix,b, 0, bufflen);
    }
    
    public synchronized void logBuffer(String prefix, byte[] b, int off, int len) {
        StringBuffer sw = new StringBuffer();        
        StringBuffer charsw = new StringBuffer();        
        if (b == null) {  // Check null buffer

            trace(prefix + " - byte array is null");
            return;
        }

        trace(prefix + " - " + len + " bytes");
        
        for (int i=0; i< len; i++) {
            int pos = off + i;
            
            if (pos >= b.length) { // Check past end of buffer 
                trace(prefix + " - " + sw.toString());
                trace(prefix + " - Attempt to read past end of buffer pos : " + pos + " bufflen: " + b.length );
                return;
            }
            sw.append(" " + ProxyServer.asHex(b[pos]));
            
            char ch = ProxyServer.asChar(b[pos]);
            
            if (ch > 127) {
                ch = '.';
            }
            else if (Character.isLetterOrDigit(ch)) {
                // allowed
            }
            else if (" !@#$%^&*()+_-=~`<>:;{[}]\"'?/\\".indexOf(ch) != -1) {
                // allowed chars 
            }
            else {
                ch = '.';
            }
            
            //charsw.append(' ');
            charsw.append(ch);

            if (i != 0 && i % 19 == 0) {
                trace(prefix + " data - " + sw.toString() + " : " + charsw.toString());
                //LOG.trace(prefix + " data - " + sw.toString());
                //LOG.trace(prefix + " data - " + charsw.toString());
                sw = new StringBuffer();
                charsw = new StringBuffer();
            }
        }
        trace(prefix + " - " + sw.toString() + ":" + charsw.toString());
        
        //LOG.trace(prefix + sw.toString());
        //LOG.trace(prefix + charsw.toString());
        
    }
	

}