
package kosoft.testproxy;

//Copyright (C) 2000 Mark O'Donohue [ mark.odonohue@gmail.com] 
//All rights reserved.
//
//This library is free software; you can redistribute it and/or modify it under the terms of the GNU 
//Lesser General Public License as published by the Free Software Foundation; either version 2.1 of 
//the License, or (at your option) any later version.

//This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
//even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
//Lesser General Public License for more details.


import java.io.*;
import java.net.*;
import java.util.*;


/**
 * This class uses the Server class to provide a multi-threaded server 
 * framework for a relatively simple proxy service.  The main() method
 * starts up the server.  The nested Proxy class implements the 
 * Server.Service interface and provides the proxy service.
 **/
public class ProxyServer {
	/** 
	 * Create a Server object, and add Proxy service objects to it to provide
	 * proxy service as specified by the command-line arguments.
	 **/

	static ProxyServer s = null;


	// This is the state for the server
	ProxyConnectionManager connectionManager; // The ConnectionManager object
	ProxyConnectionManager getConnectionManager() { return connectionManager; }
	
	Hashtable<Integer, ProxyListener> listeners; // The current services and their ports
	Hashtable<Integer, ProxyService> services;   // The current services and their ports

	ThreadGroup threadGroup;             // The threadgroup for all our threads
	PrintWriter logStream;               // Where we send our logging output to


	public ProxyServer(OutputStream logStream, int maxConnections) {
		init(logStream, maxConnections);
	}

	public void init(OutputStream logStream, int maxConnections) { 
		setLogStream(logStream);
		log("Starting server");
		threadGroup = new ThreadGroup("Server");
		connectionManager = new ProxyConnectionManager(threadGroup, maxConnections);
		connectionManager.start();
		listeners = new Hashtable<Integer, ProxyListener>();
		services = new Hashtable<Integer, ProxyService>();
	}

	/** 
	 * A public method to set the current logging stream.  Pass null
	 * to turn logging off
	 **/
	public void setLogStream(OutputStream out) {
		if (out != null) logStream = new PrintWriter(new OutputStreamWriter(out));
		else logStream = null;
	}

	protected void logWithSilent(String log) {

		if (isSilentMode()) {
			return;
		}
		log(log);
	}

	protected void logWithSilent(Object log) {

		if (isSilentMode()) {
			return;
		}
		log(log);
	}
	
	/** Write the specified string to the log */
	protected synchronized void log(String s) { 
		if (logStream != null) {
			logStream.println("[" + new Date() + "] " + s);
			logStream.flush();
		}
	}
	/** Write the specified object to the log */
	protected void log(Object o) { log(o.toString()); }

	File logDir = null;
	
	public synchronized File getLogDir() {
		if (logDir ==null) {
			logDir = new File("log");
			if (!logDir.exists()) {
				logDir.mkdirs();
			}
		}
		return logDir;
	}
	
	public void setLogDir(String dirName) {
		logDir = new File(dirName);
		if (!logDir.exists()) {
			logDir.mkdirs();
		}
	}
	
	
	/**
	 * This method makes the server start providing a new service.
	 * It runs the specified Service object on the specified port.
	 **/

	ProxyService getProxy(int port) { return services.get(port); }
	void removeProxy(int port) { services.remove(port); }

	ProxyService requireProxy(int port) { 
		ProxyService service = services.get(port);
		if (service == null) {
			throw new IllegalArgumentException("Unable to find service for port :" + port);
		}
		return service;
	}
	
	public void addService(ProxyService service)  {

		int port = service.getLocalPort();

		// Check whether a service is already on that port
		if (getProxy(port) != null) { 
			throw new IllegalArgumentException("Port " + port + " already in use.");
		}
		services.put(port, service);
		
		// Log it
		log("created proxy service on port " + service.getLocalPort() + " to "+ service.getDestName());
		//log("Starting proxy service " + service.getName() + " on port " + port);
		// Start the listener running.
		//startService(port);
	}


	/**
	 * This method makes the server stop providing a service on a port.
	 * It does not terminate any pending connections to that service, merely
	 * causes the server to stop accepting new connections
	 **/
	public void removeService(int port) {
		
		ProxyService service = requireProxy(port);

		stopService(port);

		// Close all open connections.
		service.closeAll();
		
		removeProxy(port);
		// And log it.
		log("Removed proxy service " + service.getName() + " on port " + port);
	}

	public void startService(int port)  {

		try { 
			ProxyService service = requireProxy(port);
			
			// Check whether a service is already on that port
			if (listeners.get(port) != null) { 
				throw new IllegalArgumentException("Service on port " + port + " " + service.getName() + " already started.");
			}
			
			ProxyListener listener = new ProxyListener(this, threadGroup, port, service);
	
			listeners.put(port, listener);
			log("Starting proxy service on port " + service.getLocalPort() + " to "+ service.getDestName());
			listener.start();
		}
		catch (IOException e) { 
			throw new RuntimeException(e);
		}
	}
	
	
	public void stopService(int port) {

		ProxyService service = requireProxy(port);

		ProxyListener listener = listeners.get(port);
		if (listener == null) return;
		
		// Ask the listener to stop
		listener.pleaseStop();
		// Remove it from the hashtable
		listeners.remove(port);
	}
	
	boolean silentMode = false;
	public boolean isSilentMode() { return silentMode; }
	public void setSilentMode(boolean silentMode) { this.silentMode = silentMode; }
		
	
    private static final char[] HEX_CHARS = "0123456789abcdef".toCharArray();

    static String asHex(final byte buf) {
        char[] chars = new char[2];
        chars[0] = HEX_CHARS[(buf & 0xF0) >>> 4];
        chars[1] = HEX_CHARS[buf & 0x0F];
        return new String(chars);
    }
    
    static String asHex(final byte[] buf) {
            char[] chars = new char[2 * buf.length];
            for (int i = 0; i < buf.length; ++i) {
                 chars[2 * i] = HEX_CHARS[(buf[i] & 0xF0) >>> 4];
                 chars[2 * i + 1] = HEX_CHARS[buf[i] & 0x0F];
            }
            return new String(chars);
    }

    static char asChar(final byte buf) {
             return (char)buf;
    }
    
    static String asChar(final byte[] buf) {
        StringBuffer bt = new StringBuffer();
        for (int i = 0; i < buf.length; i++) 
            bt.append((char)buf[i]);

        return new String(bt.toString());
    }
    
	/**
	 * This nested class  manages client connections for the server.  
	 * It maintains a list of current connections and enforces the
	 * maximum connection limit.  It creates a separate thread (one per
	 * server) that sits around and wait()s to be notify()'d that a connection 
	 * has terminated.  When this happens, it updates the list of connections.
	 **/
	public class ProxyConnectionManager extends Thread {
		int maxConnections;  // The maximum number of allowed connections
		Vector<ProxyConnection> connections;  // The current list of connections

		/** 
		 * Create a ConnectionManager in the specified thread group to enforce
		 * the specified maximum connection limit.  Make it a daemon thread so
		 * the interpreter won't wait around for it to exit.
		 **/
		public ProxyConnectionManager(ThreadGroup group, int maxConnections) { 
			super(group, "ConnectionManager");
			this.setDaemon(true); 
			this.maxConnections = maxConnections;
			connections = new Vector<ProxyConnection>();
			log("Starting connection manager.  Max connections: " + maxConnections);
		}

		/**
		 * This is the method that Listener objects call when they accept a
		 * connection from a client.  It either creates a Connection object 
		 * for the connection and adds it to the list of current connections,
		 * or, if the limit on connections has been reached, it closes the 
		 * connection. 
		 **/
		synchronized void addConnection(Socket s, ProxyService service) {
			// If the connection limit has been reached
			if (maxConnections != -1 && connections.size() >= maxConnections) {
				try {
					PrintWriter out = new PrintWriter(s.getOutputStream());
					// Then tell the client it is being rejected.
					out.println("Connection refused; server has reached maximum number of clients.");
					out.flush();
					s.close();
					log("Connection refused to " + s.getInetAddress().getHostAddress() + ":" + s.getPort() + ": max connections reached.");
				} 
				catch (IOException e) {log(e);}
			}
			else {  // Otherwise, if the limit has not been reached
				// Create a Connection thread to handle this connection
				ProxyConnection c = new ProxyConnection(ProxyServer.this, s, service);
				// Add it to the list of current connections
				connections.addElement(c);
				// Log this new connection
				logWithSilent("Connected to " + s.getInetAddress().getHostAddress() +
						":" + s.getPort() + " on port " + s.getLocalPort() +
						" for proxy to " + service.getName());
				// And start the Connection thread running to provide the service
				c.start();
			}
		}

		/**
		 * A Connection object calls this method just before it exits.
		 * This method uses notify() to tell the ConnectionManager thread
		 * to wake up and delete the thread that has exited.
		 **/
		public synchronized void endConnection() { this.notify(); }

		/** Change the current connection limit */
		public synchronized void setMaxConnections(int max) { maxConnections=max; }

		/** 
		 * Output the current list of connections to the specified stream.
		 * This method is used by the Control service defined below.
		 **/
		public synchronized void printConnections(PrintWriter out) {
			for(int i = 0; i < connections.size(); i++) {
				ProxyConnection c = connections.elementAt(i);
				out.println("CONNECTED TO " + 
						c.client.getInetAddress().getHostAddress() + ":" +
						c.client.getPort() + " ON PORT " + c.client.getLocalPort()+
						" FOR PROXY SERVICE TO " + c.service.getName());
			}
		}

		/**
		 * The ConnectionManager is a thread, and this is the body of that
		 * thread.  While the ConnectionManager methods above are called by other
		 * threads, this method is run in its own thread.  The job of this thread
		 * is to keep the list of connections up to date by removing connections
		 * that are no longer alive.  It uses wait() to block until notify()'d by
		 * the endConnection() method.
		 **/
		public void run() {
			while(true) {  // infinite loop
				// Check through the list of connections, removing dead ones
				for(int i = 0; i < connections.size(); i++) {
					ProxyConnection c = connections.elementAt(i);
					if (!c.isAlive()) {
						connections.removeElementAt(i);
						//log("Connection to " + c.client.getInetAddress().getHostAddress() +
						//		":" + c.client.getPort() + " closed." );
						logWithSilent("Connection to " + c.client.getInetAddress().getHostAddress() +
								":" + c.client.getPort() + " closed." 						
						+ " on port " + c.client.getLocalPort()+
						" for proxy service to " + c.service.getName());								
					}
				}
				// Now wait to be notify()'d that a connection has exited
				// When we wake up we'll check the list of connections again.
				try { synchronized(this) { this.wait(); } }
				catch(InterruptedException e) {
					try { 
						Thread.sleep(50);
					}
					catch (InterruptedException e2) {} 
				}
			}
		}
	}
	
	public static String version = "TestProxy v1.0.1";
	
	static void printVersion() {
		System.out.println(version);
	}

	public static void main(String[] args) {
		printVersion();
		try {
			s = new ProxyServer(System.out, -1);
			runMonitor(s); 
		}
		catch (Exception e) {  // Print an error message if anything goes wrong.
			System.err.println(e);
			System.err.println("Usage: java ProxyServer ");
			System.exit(1);
		}

	}


	static void runMonitor(ProxyServer s) {
		MonitorControl control = new MonitorControl(s,"test");
		try { 
			control.serve(System.in, System.out);
		}
		catch(IOException e) { } 
		System.exit(0);
	}

	static long sendDelayTime = 0;
	static long recvDelayTime = 0;


	/**
	 * This is the class that implements the proxy service.  The serve() method
	 * will be called when the client has connected.  At that point, it must
	 * establish a connection to the server, and then transfer bytes back and
	 * forth between client and server.  For symmetry, this class implements
	 * two very similar threads as anonymous classes.  One thread copies bytes
	 * from client to server, and the other copies them from server to client.
	 * The thread that invoke the serve() method creates and starts these 
	 * threads, then just sits and waits for them to exit.
	 **/


	static HashMap<String, ProxyService> proxyLookup = new HashMap<String, ProxyService>();
	static HashMap<String, ProxyService> getProxyLookup() { return proxyLookup; }

	static ProxyService getProxy(String key) { return proxyLookup.get(key); }
	static void storeProxy(ProxyService proxy) { proxyLookup.put(proxy.getKey(), proxy); }
	
	

	FilenameFilter logNameFilter = new FilenameFilter () {
		@Override
		public boolean accept(File dir, String name) {
			if (name.endsWith(".log")) {
				return true;
			}
			return false;
		}
		
	};
	
	void removeOldLogFiles() {
		File file = getLogDir();
		
		if (!file.exists()) {
			return;
		}
		
		for (File logfile : file.listFiles(logNameFilter)) {
			logfile.delete();
		}
	}
	
	boolean isLogFiles() {
		
		File file = getLogDir();
		
		if (!file.exists()) {
			return false;
		}
		
		String[] files = file.list(logNameFilter);
		if (files.length > 0 ) { 
			return true;
		}
		return false;
		
	}

	/**
	 * We check for old log files, we need to pass it once before we can create new log files.
	 * If there are old log files, then a remove of them will make this pass.
	 */
	boolean checkedForOldLogFiles = false;
	
	boolean checkForOldLogFiles() {
		
		if (checkedForOldLogFiles) {
			return false;
		}
		
		if (isLogFiles() ) {
			return true; 
		}
		
		checkedForOldLogFiles = true;
		return false;
	}
	
}

